import re
try:
    from collections.abc import MutableMapping, MutableSequence
except ImportError:
    from collections import MutableMapping, MutableSequence


class Comparison(object):
    def __init__(self, obj, **kwargs):
        self._value = obj.get('value')

    def __ne__(self, other):
        return not self.__eq__(other)

    @classmethod
    def serialize(cls):
        val = {'_t': cls.name}
        return val

class Equality(Comparison):
    name = 'equal'
    def __eq__(self, other):
        return self._value == other

class String(Comparison):
    name = 'string'
    def __eq__(self, other):
        return isinstance(other, basestring)

class Regex(Comparison):
    name = 'regex'
    def __eq__(self, other):
        return re.match(self._value, other)

class Uuid(Comparison):
    name = 'uuid'
    _re = re.compile(r'[a-f0-9]{32}')

    def __eq__(self, other):
        return self._re.match(other)

class DateHeader(Comparison):
    name = 'date'
    def __eq__(self, other):
        from email.utils import parsedate
        return bool(parsedate(other))

class ISO8601DateTime(Comparison):
    name = 'iso-8601'
    _re = re.compile(
        r'(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})'
        r'[T ](?P<hour>\d{1,2}):(?P<minute>\d{1,2})'
        r'(?::(?P<second>\d{1,2})(?:\.(?P<microsecond>\d{1,6})\d{0,6})?)?'
        r'(?P<tzinfo>Z|[+-]\d{2}:?\d{2})?$'
    )

    def __eq__(self, other):
        return bool(self._re.match(other))

class SelfReferrential(Comparison):
    name = 'self'
    def __init__(self, obj, context=None, **kwargs):
        super(SelfReferrential, self).__init__(obj, **kwargs)
        if not context:
            raise ValueError('context object required')

        self.context = context

    def __eq__(self, other):
        value = self.context.format(self._value)
        value = type(other)(value)
        return value == other


def get_comparison(value, context):
    type_field = '_t'
    if isinstance(value, MutableMapping) and type_field in value:
        _type = value[type_field]
        comparisons = {
            'equal':        Equality,
            'hash':         Equality,
            'string':       String,
            'regex':        Regex,
            'uuid':         Uuid,
            'date':         DateHeader,
            'iso-8601':     ISO8601DateTime,
            'self':         SelfReferrential,
        }

        if _type not in comparisons:
            raise ValueError('Unknown comparison type: %s' % _type)

        return comparisons[_type](value, context=context)

    return value


def compare(old, new, context):
    def compare_values(path, value, other):
        value = get_comparison(value, context)
        other = get_comparison(other, context)

        if isinstance(value, MutableMapping) and isinstance(other, MutableMapping):
            for thing in compare_dicts(path, value, other):
                yield thing

        elif isinstance(value, MutableSequence) and isinstance(other, MutableSequence):
            for thing in compare_lists(path, value, other):
                yield thing

        elif value != other:
            yield {'path': path, 'status': 'unequal', 'old': value, 'new': other}

        else:
            return

    def compare_dicts(path, value, other):
        for key in value:
            inner_path = path + [key]
            if key not in other:
                yield {'path': inner_path, 'status': 'missing', 'old': value[key]}
                continue

            for thing in compare_values(inner_path, value[key], other[key]):
                yield thing

        for key in other:
            inner_path = path + [key]
            if key not in value:
                yield {'path': inner_path, 'status': 'extra', 'new': other[key]}

    def compare_lists(path, value_list, other_list):
        for n, (value, other) in enumerate(zip(value_list, other_list)):
            inner_path = path + [n]
            for thing in compare_values(inner_path, value, other):
                yield thing

        value_len = len(value_list)
        other_len = len(other_list)
        if value_len < other_len:
            for n, other in enumerate(other_list[value_len:], start=value_len):
                yield {'path': path + [n], 'status': 'extra', 'new': other}
        elif value_len > other_len:
            for n, value in enumerate(value_list[other_len:], start=other_len):
                yield {'path': path + [n], 'status': 'missing', 'old': value}

    return list(compare_values([], old, new))



class Context(object):
    def __init__(self):
        from jinja2 import Environment
        self._env = Environment()
        self._memory = {}
        self._context = {}

    def on_response(self, test, request, response):
        self.add_test(test, request, response)

    def add_test(self, test, request, response):
        try:
            body = response.json()
        except ValueError:
            body = response.content

        context = {
            'name': test.name,
            'method': request.method,
            'url': request.url,
            'params': test.params,
            'body': body,
        }
        self._context[test.name] = context

    def format(self, string):
        template = self._env.from_string(string)
        return template.render(**self._context)

    def update_context_dict(self, value):
        from copy import deepcopy
        self._context.update(deepcopy(value))

    def update_context(self, key, value):
        self._context[key] = value



import os
import json
import hashlib
from copy import deepcopy
from requests import Request
import requests
from urlparse import urlparse
from time import time

from kevlar.checks import compare, DateHeader
from kevlar.context import Context
from kevlar.formats import JSON
from kevlar.log import LoggingHandler
from kevlar.version import VERSION_STRING

__author__ = 'Matt Long'
__copyright__ = 'Copyright 2014, Matt Long'
__license__ = 'MIT'
__version__ = VERSION_STRING
__maintainer__ = 'Matt Long'


class Test(object):
    base_url = None
    data_param_format = JSON()
    common_headers = {}

    def __init__(self, info, context, file_directory):
        self.name = info['name']
        self.method = info['verb'].upper().encode('ascii')
        self.url = self.get_full_url(context.format(info['url']))

        self.params = {}
        for k, v in info.get('params', {}).items():
            self.params[k] = context.format(str(v))

        self.files = {}
        for k, v in info.get('files', {}).items():
            v = context.format(str(v))
            self.files[k] = open(os.path.join(file_directory, v), 'rb')

        self.headers = {}
        for k, v in info.get('headers', {}).items():
            v = context.format(str(v))
            self.headers[k.encode('ascii')] = v.encode('ascii')

    def get_full_url(self, partial_url):
        u = urlparse(partial_url)

        if self.base_url and not u.scheme:
            u = urlparse('%s%s' % (self.base_url, partial_url))

        return u.geturl()

    def prepare(self):
        headers = {}
        if self.common_headers:
            headers.update(self.common_headers)
        if self.headers:
            headers.update(self.headers)

        query_params = {}
        data = {}

        if self.method in ['OPTIONS', 'GET', 'DELETE']:
            query_params.update(self.params)
        else:
            if self.files:
                # ignore data param formatting so that a multipart form will be made
                data = self.params
            else:
                data = self.data_param_format.format(self.params)
                headers['Content-Type'] = self.data_param_format.content_type

        raw_req = Request(
                method=self.method,
                url=self.url,
                headers=headers,
                data=data,
                params=query_params,
                files=self.files,
        )

        self.prepared_request = raw_req.prepare()

        return self.prepared_request

    def run(self):
        if not self.prepared_request:
            raise Exception('run() called before prepare()')

        s = requests.Session()
        resp = s.send(self.prepared_request)

        return resp.request, resp


class TestSuite(object):
    extra_test_handlers = []
    logging_class = LoggingHandler
    context_class = Context
    optional_headers = []
    global_context = None

    def __init__(self, data_directory):
        self.data_directory = data_directory
        self.extra_test_handlers = self.extra_test_handlers or []
        self.extra_modernize_functions = self.extra_modernize_functions or []
        self.context = self.context_class()
        self.optional_headers = self.optional_headers or []

        self.global_context = self.global_context or {}
        self.context.update_context_dict(self.global_context)

        internal_handlers = [self.context]
        if self.logging_class:
            internal_handlers.append(self.logging_class())

        self.handlers = internal_handlers + [handler() for handler in self.extra_test_handlers]

    def run_tests(self, tests):
        self.context.update_context('base_url', self.test_class.base_url)

        for raw_test in tests['tests']:
            test = self.test_class(raw_test, self.context, self.data_directory)

            prepared_request = test.prepare()

            for handler in self.handlers:
                if hasattr(handler, 'on_request'):
                    handler.on_request(test, prepared_request)

            request, raw_response = test.run()

            for handler in self.handlers:
                if hasattr(handler, 'on_response'):
                    handler.on_response(test, request, raw_response)

            response = self._extract_response_info(raw_response, test)
            yield test, request, response

    def _extract_response_info(self, response, test):
        try:
            body = response.json()
        except ValueError:
            raw_body = response.content
            response_path = os.path.join(self.data_directory, '%s.last-response' % test.name)
            with open(response_path, 'w') as f:
                f.write(raw_body)

            body_hash = hashlib.md5(raw_body).hexdigest()
            body = {
                '_t': 'hash',
                'value': body_hash,
            }
            print('non-JSON response was received for test "%s", using hash of contents. Response saved to %s' % (test.name, response_path))

        return {
            'status_code': response.status_code,
            'headers': dict(response.headers),
            'body': body,
        }

    @property
    def baseline_path(self):
        return os.path.join(self.data_directory, '%s_baseline.json' % self.name)

    @property
    def tests_path(self):
        return os.path.join(self.data_directory, '%s_tests.json' % self.name)

    @property
    def last_run_path(self):
        return os.path.join(self.data_directory, '%s_last_run.json' % self.name)

    def load_tests(self, update_context=True):
        with open(self.tests_path) as f:
            tests = json.load(f)

        if update_context and 'context' in tests:
            self.context.update_context_dict(tests['context'])

        #TODO validation
        return tests

    def load_baseline(self):
        with open(self.baseline_path) as f:
            baseline = json.load(f)

        #TODO validation
        return baseline

    def pretty_print(self, value):
        val = json.dumps(value, indent=2, sort_keys=True)

        # json.dumps has a bug in some versions where trailing whitespace
        # exists in dict and list lines, so we must remove it
        val = '\n'.join(map(lambda l: l.rstrip(), val.split('\n')))

        return val

    def save_baseline(self, baseline):
        with open(self.baseline_path, 'w') as f:
            f.write(self.pretty_print(baseline))

    def has_baseline(self):
        try:
            self.load_baseline()
            return True
        except IOError:
            return False

    def load_test_results(self):
        with open(self.last_run_path) as f:
            test_results = json.load(f)

        #TODO validation
        return test_results

    def save_test_results(self, test_results):
        with open(self.last_run_path, 'w') as f:
            f.write(self.pretty_print(test_results))

    def modernize_baseline(self):
        baseline = self.load_baseline()
        for name, test in baseline['tests'].items():
            self.modernize_test(test)

        self.save_baseline(baseline)

    def modernize_test(self, test):
        if 'date' in test['response']['headers']:
            test['response']['headers']['date'] = DateHeader.serialize()

        for f in self.extra_modernize_functions:
            f(test)

    def add_test_to_baseline(self, baseline, test, request, response):
        serialzied_test = {
            'name': test.name,
            'method': request.method,
            'url': request.url,
            'response': deepcopy(response),
        }
        self.modernize_test(serialzied_test)

        baseline['tests'][test.name] = serialzied_test

    def calibrate(self):
        tests = self.load_tests()

        try:
            baseline = self.load_baseline()
        except IOError:
            baseline = {
                'time': int(time()),
                'tests': {},
            }

        for (test, request, response) in self.run_tests(tests):
            if test.name in baseline['tests']:
                if test.name in []:
                    print('updating test %s in baseline, old value was:' % test.name)
                    print(json.dumps(baseline['tests'][test.name], indent=2, sort_keys=True))
                    self.add_test_to_baseline(baseline, test, request, response)
                else:
                    print('ignoring test %s since it already exists in baseline' % test.name)
            else:
                print('adding test %s to baseline' % test.name)
                self.add_test_to_baseline(baseline, test, request, response)

        self.save_baseline(baseline)

    def regress(self, add_new_tests=False):
        baseline = self.load_baseline()
        tests = self.load_tests()

        test_results = {
            'time': int(time()),
            'tests': {},
        }

        for (test, request, response) in self.run_tests(tests):
            test_result = {
                'name': test.name,
                'method': request.method,
                'url': request.url,
                'params': test.params,
                'response': response,
            }
            test_results['tests'][test.name] = test_result

            test_result['diffs'] = self.compare_test_result(baseline, test_results, test.name)
            if test_result['diffs']:
                print('---%s---' % test.name)
                for diff in test_result['diffs']:
                    print(diff)

                    if add_new_tests and diff['status'] == 'test_added':
                        self.add_test_to_baseline(baseline, test, request, response)

        self.save_test_results(test_results)
        if add_new_tests:
            self.save_baseline(baseline)

    def compare_test_result(self, baseline, new, test_name):
        diffs = []
        baseline_result = baseline['tests'].get(test_name)
        new_result = new['tests'].get(test_name)

        def serialize_path(path):
            to_string = lambda a: a if isinstance(a, basestring) else str(a)

            return '.'.join(map(to_string, path))

        optional_paths = []
        for header in self.optional_headers:
            optional_paths.append(serialize_path(['headers', header]))

        if not new_result:
            diffs.append({'status': 'test_removed', 'name': test_name})
            return diffs

        if not baseline_result:
            diffs.append({'status': 'test_added', 'name': test_name})
            return diffs

        self.context.update_context('_', new_result)

        for thing in compare(baseline_result['response'], new_result['response'], self.context):
            if thing['status'] in ['missing', 'extra']:
                if serialize_path(thing['path']) in optional_paths:
                    continue

            diffs.append(thing)

        return diffs



from setuptools import setup, find_packages

VERSION_STRING = '0.0.8'

setup(
    name="kevlar",
    packages=find_packages(),
    version = VERSION_STRING,
    author="Matt Long",
    license="MIT",
    author_email="matt@mattlong.org",
    url="https://github.com/mattlong/kevlar",
    download_url = 'https://github.com/mattlong/kevlar/tarball/' + VERSION_STRING,
    description="Kevlar will be something cool",
    install_requires=['requests>=2.4.1'],
    zip_safe=False,
    include_package_data=True,
    keywords = [],
    classifiers=[],
)



class LoggingHandler(object):
    def format_request(self, request):
        msg = []
        msg.append('{method} {url} HTTP/1.1'.format(**request.__dict__))
        for k, v in request.headers.items():
            msg.append('{0}: {1}'.format(k, v))

        return '\n'.join(msg)


    def format_response(self, response):
        msg = []
        msg.append('HTTP/1.1 {status_code} {reason}'.format(**response.__dict__))
        for k, v in response.headers.items():
            msg.append('{0}: {1}'.format(k, v))
        msg.append('')
        msg.append(response.text)
        msg.append('\n')

        return '\n'.join(msg)

    def on_request(self, test, request):
        print('===Request===')
        print(self.format_request(request))

    def on_response(self, test, request, response):
        print('\n===Response===')
        print(self.format_response(response))
